#!/usr/bin/ruby

require "rubygems"
require "json"
require "date"
require "optparse"
require "dotenv"
require "active_support/core_ext/string"

require "rollbar"
require "airbrake-ruby"

require_relative "../lib/file_manager"
require_relative "../lib/notifier"
require_relative "../lib/risk_analyzer"
require_relative "../lib/secure_file_store"


Dotenv.load()

def merge_files_hash!(output, file_hash)
  changes = output[:summary][:file_changes]
  file_hash.each do |key, value|
    changes[key] ||= {}
    changes[key].merge!(value)
  end
end

def client_factory(config, keys, additions={})
  key = keys.detect { |key| config.has_key?(key) }
  return nil if key.nil?
  require_relative "../lib/integrations/#{key}_client"
  data = config[key].merge(additions)
  "#{key.camelize}Client".constantize.new(data)
end

def scm_client(config)
  client_factory(config, ["git"])
end

def issues(config, commits)
  client_factory(config, ["jira"])
    .try(:merge_and_fetch_issues!, commits) || {}
end

def errors(config)
  client_factory(config, ["airbrake", "rollbar"])
end

def coverage(config)
  client_factory(config, ["covhura"])
end

def bugs(config, commits)
  client_factory(config, ["jira"])
    .try(:merge_and_fetch_bugs!, commits) || {}
end


def index(branches)
  config = SecureFileStore.new(ENV["SETTING_KEY"]).read_settings()
  scm_client = scm_client(config)

  branches = branches.count == 0 ?
    scm_client.remote_branches() :
    branches

  branches.each do |branch|
    commits = scm_client.commits(branch)
    file_manager = FileManager.new(branch)
    file_manager.apply!(commits, {
      issues: issues(config, commits),
      errors: errors(config),
      bugs: bugs(config, commits),
      coverage: coverage(config)
    })

    risk = RiskAnalyzer
      .new
      .analyze(file_manager, commits.last)

    Notifier
      .new
      .notify(scm_client.notification_url, commits.last, risk)
  end
end

def get_opts()
  options = {}
  OptionParser.new do |opts|
    opts.banner = "Usage: conglomerate.rb [options]"

    opts.on("-p", "--repo-path=val", String, "Repository Path") do |path|
      Dir.chdir(path)
    end

    opts.on("-v", "--verbose", "Run Verbosely") do |verbose|
      ENV["GITOLEMY_VERBOSE"] = "true"
    end

    opts.on("-c", "--compare", "Compare virtual files to Git objects") do |compare|
      ENV["GITOLEMY_COMPARE"] = "true"
    end

    opts.on("-t", "--test", "Run as test: do not store results") do |test|
      ENV["GITOLEMY_PERSIST"] = "false"
    end

    opts.on("-s", "--sync", "Sync branch even if last commit indexed") do |sync|
      ENV["GITOLEMY_SYNC"] = "true"
    end
  end.parse!
  options[:branches] = []
  options[:branches] << ARGV.first if ARGV.count > 0
  options
end


def main()
  index(get_opts()[:branches])
rescue => ex
  #handler = proc do |options|
  #  payload = options[:payload]
  #  payload["data"]["environment"] = ENV["ROLLBAR_ENV"]
  #end

  #Rollbar.configure do |config|
  #  config.access_token = ENV["ROLLBAR_API_KEY"]
  #  config.transform << handler
  #end

  #Rollbar.error(ex) if ENV["ROLLBAR_ENV"] == "production"

  #Airbrake.configure do |c|
  #  c.project_id = ENV["AIRBRAKE_PROJECT_ID"]
  #  c.project_key = ENV["AIRBRAKE_PROJECT_KEY"]
  #  c.environment = ENV["AIRBRAKE_ENV"].to_sym
  #end

  #Airbrake.notify_sync(ex) if ENV["AIRBRAKE_ENV"] == "production"

  raise ex
end

main()
