require_relative "util"
require_relative "diff"

class FileDiff
  DIFF_REGEX = /^@@ -\d+(,\d+)? \+\d+(,\d+)? @@/
  FILE_INDEX_REGEX = /index [0-9a-f]{7,40}/
  BINARY_REGEX = /^Binary files \/dev\/null and/
  DELETION_REGEX = /\[-.*?-\]/
  INSERTION_REGEX = /\{\+.*?\+\}/
  DIFF_STAT_REGEX = /^\d+\t\d+\t/
  DELETE_REGEX = /$\[-.*?-\]$/
  DELETE_OP_REGEX = /^deleted file/

  attr_accessor :operation
  attr_accessor :a_file_name
  attr_accessor :b_file_name
  attr_accessor :a_file_id
  attr_accessor :b_file_id
  attr_accessor :diffs
  attr_accessor :insertions_total
  attr_accessor :deletions_total
  attr_accessor :changes_total

  def initialize(attrs={})
    attrs.each do |key, val|
      instance_variable_set("@#{key}".to_sym, val)
    end
  end

  class << self
    def from_git(file_diff_lines)
      FileDiff.new(parse_file_diff(file_diff_lines))
    end

    # TODO: Handle different file modes like new, move, and delete.
    def parse_file_diff(file_diff_lines)
      file_names = file_diff_lines
        .first
        .sub("diff --git a/", "")
        .split(" b/")

      a_file_name = file_names.first.strip()
      b_file_name = file_names.last.strip()

      file_index = file_diff_lines
        .each_index
        .detect { |i| file_diff_lines[i].match(FILE_INDEX_REGEX) }

      if file_index.nil?
        a_file_id = nil
        b_file_id = nil
      else
        a_file_id, b_file_id = file_diff_lines[file_index]
          .sub("index ", "")
          .split(" ")
          .first
          .split("..")
      end

      if file_diff_lines.last.match(BINARY_REGEX)
        operation = :binary
      elsif (a_file_name != b_file_name)
        operation = :move
      elsif (file_diff_lines[1].match(DELETE_OP_REGEX))
        operation = :delete
      else
        operation = :change
      end

      diff_index = file_diff_lines
        .each_index
        .detect { |i| file_diff_lines[i].match(DIFF_REGEX) }

      if diff_index.nil?
        diffs = []
      else
        diffs = file_diff_lines[diff_index..-1]
          .reduce([], &fold_reducer(DIFF_REGEX))
          .flat_map { |diff_lines| Diff.from_git(diff_lines) }
      end

      insertions = diffs.map { |diff| diff.insert_count }.reduce(0, :+)
      deletions = diffs.map { |diff| diff.delete_count }.reduce(0, :+)

      {
        operation: operation,
        a_file_name: a_file_name,
        b_file_name: b_file_name,
        a_file_id: a_file_id,
        b_file_id: b_file_id,
        diffs: diffs,
        insertions_total: insertions,
        deletions_total: deletions,
        changes_total: insertions + deletions
      }
    end
  end
end
