require "active_support/core_ext/hash"

require_relative "virtual_file_system"
require_relative "file_helper"
require_relative "loggr"
require_relative "store"

class FileManager
  include FileHelper

  attr_reader :vfs

  def initialize(branch)
    @branch = branch
    @vfs = VirtualFileSystem.new()
  end

  def apply!(commits, meta)
    vfs.load!(commits, meta)
      .each do |commit|
        # TODO: Won't work for first commit...
        coverage = meta[:coverage].get_report(commit.id, @vfs.file_paths)
        errors = meta[:errors]&.get_errors!(commit) || []
        apply_commit!(commit, errors, coverage)
        Cache.index_commit(@branch, commit.id)
      end
  end

  private

  def apply_commit!(commit, errors, coverage)
    Loggr.instance.info("APPLY COMMIT: #{commit.id}")

    movements = extract_changes(commit.movements)
    changes = extract_changes(commit.changes)

    commit.trees.each { |diff_tree| @vfs.touch_tree(diff_tree, commit) }

    files = commit
      .file_diffs
      .map { |file_name, file_diff| apply_file_diff!(commit, file_diff) }

    changes.each(&method(:merge_change!))
    movements.each(&method(:merge_movement!))

    merge_coverage!(coverage) if not coverage.nil?
    errors.each { |error| apply_error!(error) }

    @vfs.save(commit)
    files.each(&:save)
  end

  def apply_file_diff!(commit, file_diff)
    file = @vfs.file_for_diff(commit, file_diff)
    file.apply_file_diff!(commit, file_diff)
    compare(commit, file.path, file.plain_text(), nil, nil) if compare?(file)
    file
  end

  def apply_error!(error)
    error[:stack_trace].each_with_index do |trace, i|
      file = trace_file(trace[:file])
      if file.nil?
        Loggr.instance.error("Trace Not Found: #{trace[:file]}:#{trace[:line]} - #{trace[:message]}")
      else
        file.merge_error!(error, trace, i) if not file.nil?
      end
    end
  rescue
    Loggr.instance.error("CANNOT APPLLY ERROR: #{error[:error_id]}")
  end

  def extract_changes(changes)
    changes.reduce({}) do |acc, (dest_file, changes)|
      changes.each do |dest_line, change|
        acc[dest_file] ||= {}
        acc[dest_file][dest_line] = @vfs[change[:from]].lines[change[:line]]
      end
      acc
    end
  end

  def merge_change!(file, file_line_changes)
    file_line_changes.each do |(line_number, line_change)|
      @vfs[file].change_line!(line_change, line_number)
    end
  end

  def merge_movement!(file, file_line_movements)
    file_line_movements.each do |(line_number, line_movement)|
      @vfs[file].move_line!(line_movement, line_number)
    end
  end

  # TODO: Technically, file lines aren't covered. Lines at commits are.
  #    1: To avoid ambiguity, maybe only view trees by commit.
  def merge_coverage!(coverage)
    coverage.each do |file, file_coverage|
      begin
        @vfs[file].merge_coverage!(file_coverage)
        @vfs[file].save()
      rescue => ex
        Loggr.instance.error("Coverage File Not Found: #{file}")
      end
    end
  end

  def trace_file(trace_file)
    match = @vfs.file_paths.detect { |app_file| trace_file.include?(app_file) }
    match.nil? ? nil : @vfs[match]
  end

  def compare?(file)
    ENV["GITOLEMY_COMPARE"] == "true" && !file.binary?
  end
end
