require "json"
require "net/http"
require "active_support/core_ext/object"

require_relative "../loggr"
require_relative "../cache"
require_relative "../store"

class JiraClient
  ISSUES_CACHE_KEY = "issues"
  STORY_TYPE = "Story"
  BUG_TYPE = "Bug"

  def initialize(config)
    @client_config = config["client_config"]
    @issue_id_rgx = issue_id_rgx
    @cached_issues = load_cached_issues()
    @business_value_custom_field = "customfield_#{config["business_value_field_id"]}"
  end

  # TODO: Store Issues as they're fetched.
  def merge_and_fetch_issues!(commits)
    issue_ids = fetch_issues(merge_issue_ids!(commits))
      .reject { |issue_id, issue| issue.empty? }
      .map { |issue_id, issue| Store::Issue::index(issue) }

    Store::Issue::cache(@cached_issues)

    issue_ids
  end

  def merge_and_fetch_bugs!(commits)
    commits
      .reduce([]) do |acc, commit|
        issue_id = parse_id(commit.subject).to_sym
        issue = issue_id.present? ? find_by_id(issue_id) : nil
        if issue.present? && issue[:issue_type] == "Bug"
          bug = {}.merge(issue)
          bug[:bug_id] = bug.delete(:issue_id)
          commit.bug_id = bug[:bug_id]
          acc << bug
        end
        acc
      end
      .map { |bug| Store::Bug::index(bug) }
  end

  private

  def client(jira_config)
    jira_config["auth_type"] = jira_config["auth_type"].to_sym
    jira_config = Hash[jira_config.map{ |key, value| [key.to_sym, value] }]
    JIRA::Client.new(jira_config)
  end

  # TODO: support many fix versions...
  def find_by_id(id)
    return @cached_issues[id] if @cached_issues.has_key?(id)

    Loggr.instance.info("FETCHING ISSUE: #{id}")
    issue = fetch_issue(id)
    {
      issue_id: id,
      status: issue.dig("fields", "status", "name"),
      assignee: extract_person(issue.dig("fields", "assignee")),
      reporter: extract_person(issue.dig("fields", "reporter")),
      issue_type: issue.dig("fields", "issuetype", "name"),
      business_value: issue.dig("fields", @business_value_custom_field) || 0.0,
      priority: issue.dig("fields", "priority", "name"),
      fix_versions: issue.dig("fields", "fixVersions").map { |fix| fix["name"] }.first,
      link: "#{@client_config["site"]}/browse/#{id}",
      summary: issue.dig("fields", "summary")
    }
  rescue
    {}
  end

  def parse_id(commit_subject)
    @issue_id_rgx
      .match(commit_subject)
      .try(:[], 0) || ""
  end

  def fetch_issues(ids)
    ids.reduce({}) do |obj, id|
      obj[id] = find_by_id(id)
      obj
    end
  end

  def merge_issue_ids!(commits)
    commits
      .map do |commit|
        commit.issue_id = parse_id(commit.subject).to_sym
        commit.issue_id
      end
      .reject { |id| id.empty? }
      .uniq
  end

  def load_cached_issues()
    Cache.read(ISSUES_CACHE_KEY, {})
      .reduce({}) do |acc, (issue_id, issue)|
        issue = issue.symbolize_keys
        issue[:issue_id] = issue_id.to_sym
        acc[issue_id.to_sym] = issue
        acc
      end
  end

  # TODO: Cache projects
  def issue_id_rgx()
    Regexp.new("(#{project_keys.join("|")})-\\d+", Regexp::IGNORECASE)
  end

  def fetch_issue(issue_id)
    request("/issue/#{issue_id}")
  end

  def extract_person(data)
    return {} if data.nil?
    {
      email: data["emailAddress"],
      name: data["displayName"]
    }
  end

  def fetch_projects()
    request("/project")
  end

  def request(uri)
    uri = URI("#{@client_config["site"]}/rest/api/2#{uri}")
    req = Net::HTTP::Get.new(uri)
    req.basic_auth(@client_config["username"], @client_config["password"])
    http = Net::HTTP.new(uri.hostname, uri.port)
    http.use_ssl = true
    JSON.parse(http.request(req).body)
  end

  def project_keys()
    fetch_projects()
      .map { |project| project["key"] }
  end
end
