require "active_support/core_ext/object"

require "amatch"
include Amatch

module Line
  JARO_SIMILARITY_THRESHOLD = 0.88
  MIN_LENGTH_SIMILARITY_THRESHOLD = 0.2
  MAX_LENGTH_SIMILARITY_THRESHOLD = 5
  COVERAGE_INVALID = nil

  TOKEN_SPLIT_REGEX = /[\s,.&|^+=*\/-]/
  MAX_MEANING_LENGTH = 400
  NOT_WHITESPACE_REGEX = /[^\s]/
  BEAUTYSPACE_REGEX = /\s/
  MEANINGLESS_KEYWORDS = {
    "" => true,
    "//" => true,
    "/*" => true,
    "*/" => true,
    "#" => true,
    "else" => true,
    "end" => true,
    "rescue" => true,
    "{" => true,
    "}" => true,
    "[" => true,
    "]" => true,
    "(" => true,
    ")" => true,
    "<" => true,
    ">" => true,
    "/>" => true
  }

  # TODO: return line_id only.
  def Line.new_line(text, file_name, line_num, commit)
    line = {
      text: text,
      coverage: COVERAGE_INVALID,
      commit: commit.id,
      issues: commit.issue_id.blank? ? [] : [commit.issue_id],
      bugs: commit.bug_id.blank? ? [] : [commit.bug_id],
      errors: [],
      revisions: []
    }
    line
  end

  # TODO: Don't move/change lines twice.
  def Line.change!(old_line, line)
    past_revisions = old_line.delete(:revisions) || []
    {
      text: line[:text],
      coverage: line[:coverage],
      commit: line[:commit],
      issues: line[:issues],
      errors: line[:errors],
      bugs: line[:bugs],
      revisions: line[:revisions] + [old_line] + past_revisions
    }
  end

  def Line.move!(old_line, line)
    Line.change!(old_line, line)
  end

  def Line.merge_error!(line, error, depth)
    return if line[:errors].any? { |line_error| line_error[:error_id] == error[:error_id] }

    line[:errors].unshift({
      error_id: error[:error_id],
      depth: depth
    })
  end

  def Line.relevant?(line)
    MEANINGLESS_KEYWORDS[line[:text].strip()].nil?
  end

  def Line.relevant_change?(line)
    return false if not relevant?(line)
    if line.has_key?(:change)
      a_line = {text: line.dig(:change, :change_text)}
      return trailing?(line, a_line) || beauty?(line, a_line)
    end
    return true
  end

  def Line.meaningless?(text)
    MEANINGLESS_KEYWORDS[text] == true ||
      text.length > MAX_MEANING_LENGTH ||
      text.split(TOKEN_SPLIT_REGEX).length <= 1
  end

  def Line.similar?(text_a, text_b)
    length_similarity = text_a.length.to_f / text_b.length.to_f
    length_similarity > MIN_LENGTH_SIMILARITY_THRESHOLD &&
      length_similarity < MAX_LENGTH_SIMILARITY_THRESHOLD &&
      text_a.jarowinkler_similar(text_b) > JARO_SIMILARITY_THRESHOLD
  end

  def Line.score(line)
    score = 0
    line[:errors].each { |error| score += 2.0 / (error[:depth] + 1) }
    score += 3 * line[:bugs].count
    score += 0.75 * line[:revisions].count
    score += 3 if line[:coverage] == 0
    score += 1.5 if line[:coverage].nil?
    score
  end

  def Line.beauty?(a_line, b_line)
    a_start = a_line[:text].index(NOT_WHITESPACE_REGEX)
    b_start = b_line[:text].index(NOT_WHITESPACE_REGEX)

    a_start == b_start &&
      a_line[:text].gsub(BEAUTYSPACE_REGEX, "") == b_line[:text].gsub(BEAUTYSPACE_REGEX, "")
  end

  def Line.trailing?(a_line, b_line)
    a_line[:text].rstrip == b_line[:text].rstrip
  end
end
