require_relative "virtual_tree"
require_relative "virtual_file"
require_relative "cache"
require_relative "store"
require "active_support/core_ext/module/delegation"

class VirtualFileSystem

  attr_reader :root

  def initialize()
    @root = VirtualTree.new({path: ""}, {})
  end

  def [](file_name)
    @root.get_file(file_name)
  end

  def file_for_diff(commit, diff)
    if diff.operation == :move
      @root.move_file(diff.a_file_name, diff.b_file_name)
    elsif diff.operation == :delete
      @root.delete_file(diff.a_file_name)
    else
      file_name = diff.b_file_name
      file = @root.get_file(file_name)
      file = @root.set_file(file_name, VirtualFile.new(file_name, diff)) if file.nil?
      file
    end
  end

  def load!(commits, meta)
    commit = commits.first
    return commits if commit.nil? || commit.cached_files.nil?
    commit.cached_trees.each do |tree_object|
      tree_id = tree_object[:object_id]
      cached_tree = Cache.read_object(tree_id)
      @root.add_tree(VirtualTree.from_json(cached_tree))
    end
    commit.cached_files.each do |file_object|
      file_id = file_object[:object_id]
      cached_file = Cache.read_object(file_id)
      if not cached_file.nil?
        @root.set_file(file_object[:path], VirtualFile.from_json(file_id, cached_file))
      end
    end
    sync_meta!(commit, meta)
    commits
  end

  def touch_tree(diff_tree, commit)
    @root.touch(diff_tree, commit)
  end

  def save(commit)
    prune(commit)
    @root.save(commit)
  end

  def file_paths
    root
      .all_files()
      .map { |file| file.path }
  end

  private

  def prune(commit)
    commit
      .trees
      .reverse()
      .each { |diff_tree| @root.prune(diff_tree, commit) }
  end

  def sync_meta!(commit, meta)
    meta[:errors] && meta[:errors].sync!(commit)
  end
end
